#######################################################################
###### Plugin Software 'BatchLR' was Designed and Developed by
###### BDLS Informatics, Product Innovation.
###### Authored by Joshua Luthy, and Ian Taylor
###### This template was last updated in August 2019
###### The underlying R package 'Batchelor', and Mutual Nearest
###### Neighbors algorithm on which that package is based, was
###### developed by: Aaron Lun, Cancer Research UK Cambridge Institute
#######################################################################
#######################################################################
##         LOAD INPUTS  & FUNCTIONS     ##
#######################################################################
sessionInfo()

setwd("SG_OUTPUT_DIRECTORY_PATH")
outs <- "SG_OUTPUT_DIRECTORY_PATH"
importCorrectedGenes <- CORRECT_RAW
parBatchCorrection <- FJ_BATCH_COR
parCorrelationAnalysis <- FJ_CA
parHarmonize <- FJ_HARMONIZE
knn <- KNN

StopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(outs, "/Error_Message_from_R.txt"))
  stop(message, ...)
}
#######################################################################
###### Start package import or installations
#######################################################################
# Install function for packages
loadPackages <-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    install.packages(need, repos="https://cran.r-project.org")
    lapply(need,require,character.only=TRUE)
  }
}

loadBioc<-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    BiocManager::install(need)
    lapply(need,require,character.only=TRUE)
  }
}
tryCatch({
  loadPackages(c('dplyr','readr','BiocManager','Seurat','tibble','cowsay','data.table','cowplot','devtools'))
}, error = function(e) {
  message <- "There was an error with installing R packages. Please see How-To PDF and try to manually install."
  StopAndReturnError(message = message)
}, warning = function(w){
  loadPackages(c('dplyr','readr','BiocManager','Seurat','tibble','cowsay','data.table','cowplot','devtools'))
  writeLines("loaded R packages")
})

tryCatch({
  loadBioc(c('batchelor','scater','SingleCellExperiment','scran'))
}, error = function(e) {
  message <- "There was an error with installing R packages. Please see How-To PDF and try to manually install."
  StopAndReturnError(message = message)
}, warning = function(w) {
  loadBioc(c('batchelor','scater','SingleCellExperiment','scran'))
  writeLines("loaded R packages")
})
# The below should work for R-4.1 and later.
Sys.setenv("PKG_CXXFLAGS"="-std=c++11")
if(!require("harmony",character.only=TRUE)){
  suppressMessages(suppressWarnings(devtools::install_github("immunogenomics/harmony")))
}

tryCatch({
  library("harmony")
}, error = function(e) {
  message <- "BatchLR was not able to run. There is an error loading the Harmony R package. Please see How-To PDF and try to manually install."
  StopAndReturnError(message = message)
})

#######################################################################
###### Start the File Import
#######################################################################
# Read CSV from SeqGeq
importedCSVfile <- fread(file= "SG_SAMPLE_DATA_PATH",
                         sep = ",", header = TRUE, check.names = FALSE)
featureAnnotationsFrame <- data.frame(importedCSVfile$Gene)
featureAnnos <- as.character(importedCSVfile$Gene)
cellAnnotationsFrame <- data.frame(colnames(importedCSVfile))

if(any(names(importedCSVfile) == "Gene")) {
  cellAnnos <- as.character(colnames(importedCSVfile))
  cellAnnos <- cellAnnos[cellAnnos != "Gene"]

  cellAnnotationsFrame <- data.frame(colnames(importedCSVfile))
  rowsToKeep <- importedCSVfile[, which(names(importedCSVfile) != "Gene")]
  cellAnnotationsFrame <- data.frame(cellAnnotationsFrame[rowsToKeep,])
  colnames(cellAnnotationsFrame) <- paste("CellId")

}
rm(rowsToKeep)

importedCSVfile <- as.data.frame(importedCSVfile)
rownames(importedCSVfile) <- featureAnnos
importedCSVfile <- dplyr::select(importedCSVfile, -Gene)

batchFrame <- importedCSVfile["SampleID",]

# Some data has numbers as cellIDs only, append a string to cellIds, remove on export.
colnames(importedCSVfile) <- paste("AppID", colnames(importedCSVfile), sep="_")
#TODO can remove this next line if batchFrame no longer needed.

tryCatch({
# Check if there are SampleIDs to use as SCE Batches
batchFrameCheck <- as.factor(batchFrame)
len <- levels(batchFrameCheck)

if(length(len) < 2){
  print("Only one SampleID found, please create the concatenated file again.")
  StopAndReturnError("Too few 'SampleID' values found. Try concatenating \n two or more populations in SeqGeq and run \n BatchLR on this new file.", call. = FALSE)
}
}, error = function(e) {
  print("Can't check how many Samples are present.")
})


if(parHarmonize){
  batchFrame <- t(batchFrame)
}

if (parBatchCorrection) {
  say(what="Starting MNN Batch Correction", by= "owl")

# #################################################################
# #   Performing Batch Correction on two or more batches
# #################################################################
# # Performing fastMNN batch correction on the two imported SeqGeq populations.
# # directly for downstream analyses such as clustering and visualization.
# TODO: Give a Nice error message if Waning: "tied distances detected in nearest-neighbor calculation" is thrown.
  newDF <- data.frame(t(importedCSVfile))
  splitDF <- split(newDF, newDF$SampleID)
  # now order them from high to low for merging steps
  splitDF <- splitDF[order(sapply(splitDF, NROW), decreasing = T)]
  for(i in seq_along(splitDF)) {
    splitDF[[i]] <- subset(splitDF[[i]], select = -c(SampleID))
  }
  # Now transpose them to params in cloumns for Seurat
  for(i in seq_along(splitDF)){
    splitDF[[i]] <- t(splitDF[[i]])
  }

  output <- vector("list", length(splitDF))

  # create a list SingleCellExperiment objects
  for (i in seq_along(splitDF)) {
    output[[i]] <- SingleCellExperiment(list(counts=as.matrix(splitDF[[i]]),
                                             logcounts = log2(as.matrix(splitDF[[i]]) + 1)
    ))
    # colData=cellAnnotationsFrame,
    # rowData=featureAnnotationsFrame)
  }
  ########
  ########
  # # Get the counts of columns in each matrix
  # batchCounts <- sapply(output, NCOL)
  # batchCounts <- as.data.frame(batchCounts)
  #
  # # Order the rows from high to low
  # batchCounts$mergeOrder <- seq_len(nrow(batchCounts))
  # batchCounts <- batchCounts[order(-batchCounts$batchCounts),]
  # mergeOrdervec <- batchCounts$mergeOrder
  ########
  ########

  #################################################################
  #   Performing Batch Correction on two or more batches
  #################################################################
  # Performing fastMNN batch correction on the two imported SeqGeq populations.
  # directly for downstream analyses such as clustering and visualization.
  # If possible, authors recommend using the output of multiBatchNorm() as input to fastMNN.
  # This will equalize coverage on the count level before the log-transformation, which is a more
  # accurate rescaling than cosine normalization on the log-values.
  batchCorrect <- function(inputList,knn){
    result <- tryCatch(withCallingHandlers(
    {
      say("Normalize Batches and Perform Batch Correction", "cat")
      mnn.out <<- multiBatchNorm(inputList, norm.args=list(use_altexps=FALSE)) %>%
        fastMNN(., k = knn)
    },
      error = function(e){
        print(paste0("Error when normalizing batches, performing batch correction without normalizing batches: ",e))
        mnn.out <<- fastMNN(inputList, k = knn)
      }
    ))
    return(result)
  }

  batchCorrect(inputList = output, knn = knn)

  # normed <- multiBatchNorm(output, norm.args=list(use_altexps=FALSE))
  # mnn.out <- batchelor::fastMNN(normed, k = )

  # Display string of corrected values
str(reducedDim(mnn.out, "corrected"))

#################################################################
##### Viewing the Batch Corrected Pops
#################################################################
# Each column of mnn.out corresponds to a cell in one of the batches,
# while each row corresponds to an input gene in chosen. The corrected
# matrix in the reducedDims slot contains the low-dimensional corrected coordinates for all cells.
batches <- list(mnn.out)
head(batches)
dim(reducedDim(mnn.out, "corrected"))
reducedDims(mnn.out)

# Using an (run length encoding) Rle for pretty-printing of batch IDs
# (as all cells from the same batch are consecutive).

print(Rle(mnn.out$batch))

#################################################################
##### Effect of correction with Diagnostics
#################################################################
# One useful diagnostic is the proportion of variance within each batch
# that is lost during MNN correction. Specifically, this refers to the
# within-batch variance that is removed during orthogonalization with
# respect to the average correction vector at each merge step. This is
# returned via the lost.var field in the metadata of mnn.out, which contains
# a matrix of the variance lost in each batch (column) at each merge step (row).

Lost_Variance <- metadata(mnn.out)$merge.info$lost.var

writeLines(" Difference in variance lost between batches:  ")
print(Lost_Variance)
writeLines(" Large values, and major differences between batches indicates a loss of true biological signal. ")
writeLines(" Differences between the batch effect and biological signal are assumed to be orthagonal.(Haghverdi et al. 2018) ")

## EG of "good values", low/similar losses:
##             [,1]       [,2]
## [1,] 0.02094553 0.03466084

# Large proportions of lost variance suggest that correction is removing genuine
# biological heterogeneity. This would occur due to violations of the assumption of
# orthogonality between the batch effect and the biological subspace (Haghverdi et al. 2018).

# TODO What is considered a small variance here? TODO
# TODO Print these values out for each batch, in a sensible way TODO

#################################################################
#   Plotting two Batch Corrected Populations  #
#################################################################
mnnPCs <-  plotReducedDim(mnn.out, colour_by="batch", dimred = "corrected", ncomponents = 5)
  ggsave(paste0(outs, "MNN_components.png"), plot = mnnPCs,  width = 6, height = 6, device = "png")

#################################################################
#   Create a Header for the first 25 PC  #
#################################################################

MNNpcHeader <- data.frame()
for(i in 1:25){
  de_i <- list(paste("MNNpc","SG_RUNID",i))
  MNNpcHeader <- rbind(MNNpcHeader,de_i,stringsAsFactors=FALSE,head(NULL))
}
MNNpcHeaderParams <- as.data.frame(MNNpcHeader, row.names=NULL)
write.table(MNNpcHeaderParams,paste0("MNN-Params_","SG_RUNID",".csv"), quote=FALSE, row.names=FALSE, col.names=FALSE)
#################################################################
#   Exporting the corrected values to a .csv  #
#################################################################

pcMNNout <- as_tibble(reducedDim(mnn.out, "corrected"), .name_repair = "minimal")
MNNpcOut.tmp <- pcMNNout[,1:25]
colnames(MNNpcOut.tmp) <- paste(MNNpcHeader$X.MNNpc.SG_RUNID.1)
MNNpcOut <- MNNpcOut.tmp %>% tibble::add_column(cellAnnotationsFrame) %>%
  dplyr::select(CellId, everything())

#cAF3 <- data.frame(cellAnnotationsFrame)
#rownames(MNNpcOut) <- paste(cAF3$CellID)
# Use write_csv to prevent rownames being written to the csv file.
  if(parBatchCorrection && !parCorrelationAnalysis && !parHarmonize){
    write_csv(MNNpcOut, paste0("BatchLR_PCA_","SG_RUNID",".csv"))
  } else { say(what="Completed MNN Batch Correction", by = "cat") }

  if (importCorrectedGenes){
    #################################################################
    #   Exporting the Corrected Per Gene Values to a .csv  #
    #################################################################
    # In the 'assays' slot is a reconstructed matrix containing the
    # low-rank reconstruction of the expression matrix. This can be
    # interpreted as per-gene corrected log-expression values (after
    # cosine normalization, if cos.norm = TRUE) but should not be used
    # for quantitative analyses. This has number of rows equal to the
    # number of input genes if subset.row = NULL or if correct.all = TRUE,
    # otherwise each row corresponds to a gene in subset.row.
    center_mean <- function(x) {
      ones = rep(1, nrow(x))
      x_mean = ones %*% t(colMeans(x))
      x - x_mean
    }

    # Working with tibble
    #perGeneCorMatrix.tmp <- as_tibble(assays(mnn.out), .name_repair = "minimal") # This is per Gene
    #perGeneCorMatrix.tmp <- perGeneCorMatrix.tmp[,-c(1:2)] # removes two additional columns that are not needed
    #batchFrame2 <- data.frame(t(batchFrame))
    #perGeneCorMatrix <- as_tibble(t(perGeneCorMatrix.tmp), .name_repair = "minimal") %>% tibble::add_column(cellAnnotationsFrame) %>%
    #  tibble::add_column(batchFrame2) %>% dplyr::select(CellId, everything()) %>% remove_rownames() %>%
    #  column_to_rownames(var="CellId")
    #perGeneCorMatrix <-  center_mean(perGeneCorMatrix) # Mean center the data
    #perGeneCorMatrix <- ((perGeneCorMatrix)^2)*1000
    #
    perGeneCorMatrix <- as.data.frame(assays(mnn.out)) # This is per Gene
    perGeneCorMatrix <- perGeneCorMatrix[,-c(1:2)] # removes two additional columns that are not needed
    perGeneCorMatrix <-  center_mean(perGeneCorMatrix) # Mean center the data
    # perGeneCorMatrix <- ((perGeneCorMatrix)^2)*20000
    perGeneCorMatrix[nrow(perGeneCorMatrix)+1,] <- batchFrame
    perGeneCorMatrix <- as_tibble(t(perGeneCorMatrix))
    perGeneCorMatrix <- perGeneCorMatrix %>% tibble::add_column(cellAnnotationsFrame)
    perGeneCorMatrix <- as.data.frame(perGeneCorMatrix)
    perGeneCorMatrix <- perGeneCorMatrix %>% dplyr::select(CellId, everything())
    write.csv(perGeneCorMatrix, paste0("MNN_Corrected_Matrix_","SG_RUNID",".csv"), row.names = FALSE) # export the log values of corrected matrix
  }
}

###############################################################
#  Creating Seurat Object(s) for running reciprocal PCA   #
###############################################################
if (parCorrelationAnalysis) {
  say(what="Starting rPCA Correlation Analysis", by= "cow")
  # library(Seurat)
  # library(SeuratWrappers) # can use this if we want to use Seurat+Batchelor for batch correction as well.
  SGparams <- 2000 #JL.this value may come from SG (use up to 2000 for 3-5 data sets and 3000 for more than 5)

  # BatchLR exports as PIR (Parameters in Rows)
  # Transpose the CSV so that the column 'SampleID' can be used for formatting prior to making a Seurat Object
  newDF <- data.frame(t(importedCSVfile))
  splitDF <- split(newDF, newDF$SampleID)
  # now order them from high to low for merging steps
  splitDF <- splitDF[order(sapply(splitDF, NROW), decreasing = T)]
  # now lets name these by sample name from geqzip( getSamples) or from FJServer and my Noir package...?
  # namesToAdd <- getSamples(urlSams) # from Noir
  # namesToAdd <- c("Ctrl.csv", "BACT.csv", "Day4-BACT") # This list should come from SeqGeq Java
  # alternate to nzchar:is a fast way to find out if elements of a character vector are non-empty strings? ideally would use !nzchar
  # names(splitDF)[nzchar(names(splitDF))] <- namesToAdd # ideally would use !nzchar
  # Now remove the SampleID column before making Seurat Objects
  for(i in seq_along(splitDF)) {
    splitDF[[i]] <- subset(splitDF[[i]], select = -c(SampleID))
  }
  # Now transpose them to params in cloumns for Seurat
  for(i in seq_along(splitDF)){
    splitDF[[i]] <- t(splitDF[[i]])
  }
  ###################################
  ## Create List of Seurat Objects ##
  ###################################
  # for (i in seq_along(splitDF)) {            # 2. sequence
  #   output[[i]] <- CreateSeuratObject(splitDF[[i]]) #, project = namesToAdd[i])      # 3. body
  # }
  output <<- vector("list", length(splitDF))
  for (i in seq_along(splitDF)) {            # 2. sequence
    output[[i]] <- CreateSeuratObject(splitDF[[i]]) #, project = namesToAdd[i])      # 3. body
  }

  # After acquiring the data, we first perform standard normalization and variable feature selection.
  for (i in seq_along(output)) {
    output[[i]] <- NormalizeData(output[[i]], verbose = FALSE)
  }

  for (i in seq_along(output)) {
    output[[i]] <- FindVariableFeatures(output[[i]], nfeatures = 2000, verbose = FALSE)
  }

  # Next, select features for downstream integration, and run PCA on each object in the list,
  # which is required for running the reciprocal PCA workflow.
  myFeatures <- SelectIntegrationFeatures(object.list = output, nfeatures = 2000)
  output <- lapply(X = output, FUN = function(x) {
    x <- ScaleData(x, features = myFeatures, verbose = FALSE)
    x <- RunPCA(x, features = myFeatures, verbose = FALSE)
  })

  # Next, identify anchors and integrate the datasets. Commands are identical to the standard workflow, but make sure
  # to set reduction = 'rpca':
  anchors <- FindIntegrationAnchors(object.list = output, reference = NULL, reduction = "rpca",
                                    dims = 1:30)
  output.integrated <- IntegrateData(anchorset = anchors, dims = 1:40) # Consider using "preserve.order=TRUE"
  # Now proceed with downstream analysis (i.e. visualization, clustering) on the integrated dataset.
  # Commands are identical to the standard workflow.
  output.integrated <- ScaleData(output.integrated, verbose = FALSE)
  output.integrated <- RunPCA(output.integrated, verbose = FALSE, reduction.key = paste0("rPCApc_","SG_RUNID_"))

  # output.integrated <- RunUMAP(output.integrated, dims = 1:30)
  # DimPlot(output.integrated, group.by = "orig.ident")

  ################################################################
  #   Create a Header for the first 25 PC  #
  #################################################################

  rPCApcHeader <- data.frame()
  for(i in 1:25){
    de_i <- list(paste("rPCApc","SG_RUNID",i))
    rPCApcHeader <- rbind(rPCApcHeader,de_i,stringsAsFactors=FALSE,head(NULL))
  }
  rPCApcHeaderParams <- as.data.frame(rPCApcHeader, row.names=NULL)
  write.table(rPCApcHeaderParams,paste0("rPCA-Params_","SG_RUNID",".csv"), quote=FALSE, row.names=FALSE, col.names=FALSE)
  #################################################################
  #   Exporting the corrected values from recipricalPCA to a .csv #
  #################################################################
  rPCA <- as_tibble(output.integrated@reductions[["pca"]]@cell.embeddings, .name_repair = "minimal")
  MCAout.tmp <- rPCA[,1:25]
  colnames(MCAout.tmp) <- paste(rPCApcHeader$X.rPCApc.SG_RUNID.1)

  MCAout <- MCAout.tmp %>% tibble::add_column(cellAnnotationsFrame) %>%
    dplyr::select(CellId, everything())

  #cAFWunTwo <- data.frame(cellAnnotationsFrame)
  #rownames(MCAout) <- paste(cAFWunTwo$CellID)

  # Use write_csv to prevent rownames being written to the csv file.
  if(parCorrelationAnalysis && !parBatchCorrection && !parHarmonize){
  write_csv(MCAout, paste0("BatchLR_PCA_", "SG_RUNID",".csv"))
}

  if (importCorrectedGenes){
    center_mean <- function(x) {
      ones = rep(1, nrow(x))
      x_mean = ones %*% t(colMeans(x))
      x - x_mean
    }
    correctedScaledData <- data.frame(output.integrated@assays$integrated@scale.data)
    # correctedScaledData <- ((correctedScaledData)^2)*10000
    correctedScaledData[nrow(correctedScaledData)+1,] <- batchFrame
    correctedScaledData <- as_tibble(t(correctedScaledData))
    correctedScaledData <- correctedScaledData %>% tibble::add_column(cellAnnotationsFrame) #%>% remove_rownames() %>% column_to_rownames(var = "CellId")
    correctedScaledData <- as.data.frame(correctedScaledData)
    correctedScaledData <- correctedScaledData %>% dplyr::select(CellId, everything())
    #colnames(correctedScaledData) <- paste(cellAnnos) # return the original names
    #correctedScaledData <- as.data.frame(t(correctedScaledData))
    write_csv(correctedScaledData, paste0("rPCA_Corrected_Matrix_","SG_RUNID",".csv"))
  }
}

if(parHarmonize){
  say("Performing batch correction with harmony in R", by = "spider")
  # remove SampleID from batch correction matrix
  importedCSVfile.tmp <- importedCSVfile
  harmonyMatrix <-  importedCSVfile.tmp[!(row.names(importedCSVfile.tmp) %in% "SampleID"), ]
  rm(importedCSVfile.tmp)
  seuratSeqGeq <- CreateSeuratObject(harmonyMatrix)
  seuratSeqGeq@meta.data$batchID <- as.factor(batchFrame)
  seuratSeqGeq <- NormalizeData(seuratSeqGeq, normalization.method = "LogNormalize")
  seuratSeqGeq <- FindVariableFeatures(seuratSeqGeq, nfeatures = 2000, verbose = FALSE)
  #varFeats <- seuratSeqGeq@assays[["RNA"]]@var.features # needed for SG or only FJ?
  #Scale the data
  all.genes <- rownames(seuratSeqGeq)
  all.genes <- all.genes[all.genes != "SampleID"]
  seuratSeqGeq <- ScaleData(seuratSeqGeq, features = all.genes)
  seuratSeqGeq <- RunPCA(seuratSeqGeq, features = VariableFeatures(seuratSeqGeq))
  #RunHarmony correction and print out png of iterations showing convergence
  seuratSeqGeq <- RunHarmony(seuratSeqGeq, "batchID",verbose = TRUE, plot_convergence = TRUE, reduction.save = "harmony")
  ################################################
  ## Optional plots to see pre batch correction ##
  ################################################
  #options(repr.plot.height = 5, repr.plot.width = 12)
  #p1pre <- DimPlot(object = seuratSeqGeq, reduction = "pca", pt.size = .1, group.by = "batchID")
  #p2pre <- VlnPlot(object = seuratSeqGeq, features = "PC_1", group.by = "batchID", pt.size = .1)
  #png(file = "PreCorrection.png")
  #plot_grid(p1pre,p2pre)
  #dev.off()
  #################################################
  ## Optional plots to see post batch correction ##
  #################################################
  #options(repr.plot.height = 5, repr.plot.width = 12)
  #p1post <- DimPlot(object = seuratSeqGeq, reduction = "harmony", pt.size = .1, group.by = "batchID")
  #p2post <- VlnPlot(object = seuratSeqGeq, features = "harmony_1", group.by = "batchID", pt.size = .1)
  #png(file="PostCorrection.png")
  #plot_grid(p1post,p2post)
  #dev.off()

  HarmonyPcHeader <- data.frame()
  for(i in 1:25){
    de_i <- list(paste("harmony","SG_RUNID",i))
    HarmonyPcHeader <- rbind(HarmonyPcHeader,de_i,stringsAsFactors=FALSE,head(NULL))
  }
  HarmonyPcHeaderParams <- as.data.frame(HarmonyPcHeader, row.names=NULL)
  write.table(HarmonyPcHeaderParams,paste0("Harmony-Params_","SG_RUNID",".csv"), quote=FALSE, row.names=FALSE, col.names=FALSE)
  #################################################################
  #   Exporting the corrected values from recipricalPCA to a .csv #
  #################################################################
  hrmnyPCA <- as_tibble(seuratSeqGeq@reductions[["harmony"]]@cell.embeddings, .name_repair = "minimal")
  hrmnyPCA.tmp <- hrmnyPCA[,1:25]
  colnames(hrmnyPCA.tmp) <- paste(HarmonyPcHeader$X.harmony.SG_RUNID.1)

  HrmnyOut <- hrmnyPCA.tmp %>% tibble::add_column(cellAnnotationsFrame) %>%
    dplyr::select(CellId, everything())

  #cAFWunTwo <- data.frame(cellAnnotationsFrame)
  #rownames(MCAout) <- paste(cAFWunTwo$CellID)

  # Use write_csv to prevent rownames being written to the csv file.
  if(parHarmonize && !parCorrelationAnalysis && !parBatchCorrection){
    write_csv(HrmnyOut, paste0("BatchLR_PCA_", "SG_RUNID",".csv"))
  }else{say("Oops, failed to export the file of PCs. Look in output folder or re-run plugin", by="spider")}
}

## This will create one csv with both sets of principal compenents.
#if(parBatchCorrection && parCorrelationAnalysis) {
#   AllOut <- select(batchlrCCA, -c(CellId))
#  allPCs <- cbind(batchlrMNN, batchlrCCA)
#  write_csv(allPCs, paste0("BatchLR_PCA_", "SG_RUNID",".csv"))
#} else {
#  writeLines(" Only One Batch Correction Analysis Implemented, Returning One Parameter Set")
#}
say("time")